"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.convertTSType = void 0;
const t = __importStar(require("@babel/types"));
const convertTSEntityName_1 = require("./convertTSEntityName");
const convertFunctionTypeAnnotation_1 = require("./convertFunctionTypeAnnotation");
const baseNodeProps_1 = require("../utils/baseNodeProps");
const convertTSTypeLiteral_1 = require("./convertTSTypeLiteral");
function convertTSType(node) {
    if (t.isTSAnyKeyword(node)) {
        return t.anyTypeAnnotation();
    }
    if (t.isTSParenthesizedType(node)) {
        return convertTSType(node.typeAnnotation);
    }
    if (t.isTSArrayType(node)) {
        const elementType = Object.assign(Object.assign({}, convertTSType(node.elementType)), baseNodeProps_1.baseNodeProps(node.elementType));
        return t.arrayTypeAnnotation(
        // t.isUnionTypeAnnotation(elementType) || t.isIntersectionTypeAnnotation(elementType)
        //   ? t.parenthesizedType(elementType)
        //   :
        elementType);
    }
    if (t.isTSUndefinedKeyword(node)) {
        return t.voidTypeAnnotation();
    }
    if (t.isTSBooleanKeyword(node)) {
        return t.booleanTypeAnnotation();
    }
    if (t.isTSLiteralType(node)) {
        if (t.isBooleanLiteral(node.literal)) {
            return t.booleanLiteralTypeAnnotation(node.literal.value);
        }
        else if (t.isNumericLiteral(node.literal)) {
            return t.numberLiteralTypeAnnotation(node.literal.value);
        }
        else if (t.isStringLiteral(node.literal)) {
            return t.stringLiteralTypeAnnotation(node.literal.value);
        }
        else if (t.isUnaryExpression(node.literal)) {
            // todo: @babel/types node.literal is not suposed to be a unray expression
            const exp = node.literal;
            if (t.isNumericLiteral(exp.argument) && exp.operator === '-') {
                return t.numberLiteralTypeAnnotation(-exp.argument.value);
            }
            else {
                throw new Error('not implemented');
            }
        }
        else {
            throw new Error('not implemented');
        }
    }
    if (t.isTSNeverKeyword(node)) {
        return t.emptyTypeAnnotation();
    }
    if (t.isTSTypeReference(node)) {
        const typeName = convertTSEntityName_1.convertTSEntityName(node.typeName);
        let flowTypeParameters = null;
        if (node.typeParameters) {
            const flowParams = node.typeParameters.params.map(p => (Object.assign(Object.assign({}, convertTSType(p)), baseNodeProps_1.baseNodeProps(p))));
            flowTypeParameters = t.typeParameterInstantiation(flowParams);
        }
        if (t.isIdentifier(typeName) && typeName.name === 'Readonly') {
            return t.genericTypeAnnotation(t.identifier('$ReadOnly'), flowTypeParameters);
        }
        else if (t.isIdentifier(typeName) && typeName.name === 'ReadonlyArray') {
            return t.genericTypeAnnotation(t.identifier('$ReadOnlyArray'), flowTypeParameters);
        }
        else if (t.isIdentifier(typeName) && typeName.name === 'Partial') {
            return t.genericTypeAnnotation(t.identifier('$Shape'), flowTypeParameters);
        }
        else if (t.isIdentifier(typeName) && typeName.name === 'NonNullable') {
            return t.genericTypeAnnotation(t.identifier('$NonMaybeType'), flowTypeParameters);
        }
        else if (t.isIdentifier(typeName) && typeName.name === 'ReadonlySet') {
            return t.genericTypeAnnotation(t.identifier('$ReadOnlySet'), flowTypeParameters);
        }
        else if (t.isIdentifier(typeName) && typeName.name === 'ReadonlyMap') {
            return t.genericTypeAnnotation(t.identifier('$ReadOnlyMap'), flowTypeParameters);
        }
        else if (t.isIdentifier(typeName) && typeName.name === 'ReturnType') {
            const params = flowTypeParameters === null
                ? [t.anyTypeAnnotation()]
                : flowTypeParameters.params;
            return t.genericTypeAnnotation(t.identifier('$Call'), t.typeParameterInstantiation([
                //<R>((...args: any[]) => R) => R,
                t.functionTypeAnnotation(t.typeParameterDeclaration([
                    Object.assign(Object.assign({}, t.typeParameter()), { name: 'R' }),
                ]), [
                    t.functionTypeParam(null, t.functionTypeAnnotation(null, [], t.functionTypeParam(t.identifier('args'), t.arrayTypeAnnotation(t.anyTypeAnnotation())), t.genericTypeAnnotation(t.identifier('R')))),
                ], null, t.genericTypeAnnotation(t.identifier('R'))),
                ...params,
            ]));
        }
        else if (t.isIdentifier(typeName) && typeName.name === 'Record') {
            if (flowTypeParameters === null ||
                flowTypeParameters.params.length !== 2) {
                // todo: warn
                return t.genericTypeAnnotation(t.identifier('Record'), flowTypeParameters);
            }
            return t.objectTypeAnnotation([], [
                t.objectTypeIndexer(t.identifier('key'), flowTypeParameters.params[0], flowTypeParameters.params[1]),
            ], [], []);
        }
        else if (t.isIdentifier(typeName) &&
            typeName.name === 'Omit' &&
            flowTypeParameters &&
            flowTypeParameters.params.length === 2 &&
            (t.isStringLiteralTypeAnnotation(flowTypeParameters.params[1]) ||
                (t.isUnionTypeAnnotation(flowTypeParameters.params[1]) &&
                    flowTypeParameters.params[1].types.every(type => t.isStringLiteralTypeAnnotation(type))))) {
            // special case when Omit can be directly replaced by `$Rest`
            return t.genericTypeAnnotation(t.identifier('$Rest'), t.typeParameterInstantiation([
                flowTypeParameters.params[0],
                t.objectTypeAnnotation((t.isStringLiteralTypeAnnotation(flowTypeParameters.params[1])
                    ? [flowTypeParameters.params[1]]
                    : flowTypeParameters.params[1]
                        .types).map(type => t.objectTypeProperty(t.identifier(type.value), t.anyTypeAnnotation())), [], [], []),
            ]));
        }
        else {
            return t.genericTypeAnnotation(typeName, flowTypeParameters);
        }
    }
    if (t.isTSIndexedAccessType(node)) {
        if (t.isTSLiteralType(node.indexType) &&
            t.isStringLiteral(node.indexType.literal)) {
            node.indexType;
            return t.genericTypeAnnotation(t.identifier('$PropertyType'), t.typeParameterInstantiation([
                convertTSType(node.objectType),
                convertTSType(node.indexType),
            ]));
        }
        else {
            return t.genericTypeAnnotation(t.identifier('$ElementType'), t.typeParameterInstantiation([
                convertTSType(node.objectType),
                convertTSType(node.indexType),
            ]));
        }
    }
    if (t.isTSIntersectionType(node)) {
        const tsTypes = node.types;
        return t.intersectionTypeAnnotation(tsTypes.map(v => {
            const flowType = convertTSType(v);
            // if (t.isTSFunctionType(flowType)) {
            //   flowType = t.tsParenthesizedType(flowType);
            // }
            return Object.assign(Object.assign({}, flowType), baseNodeProps_1.baseNodeProps(v));
        }));
    }
    if (t.isTSUnknownKeyword(node)) {
        return t.mixedTypeAnnotation();
    }
    if (t.isTSNullKeyword(node)) {
        return t.nullLiteralTypeAnnotation();
    }
    if (t.isTSNumberKeyword(node)) {
        return t.numberTypeAnnotation();
    }
    if (t.isTSMappedType(node)) {
        return t.objectTypeAnnotation([], [
            t.objectTypeIndexer(t.identifier(node.typeParameter.name), node.typeParameter.constraint
                ? convertTSType(node.typeParameter.constraint)
                : t.anyTypeAnnotation(), node.typeAnnotation
                ? convertTSType(node.typeAnnotation)
                : t.anyTypeAnnotation(), node.readonly ? t.variance('plus') : null),
        ], [], []);
    }
    if (t.isTSTypeLiteral(node)) {
        return convertTSTypeLiteral_1.convertTSTypeLiteral(node);
    }
    if (t.isTSStringKeyword(node)) {
        return t.stringTypeAnnotation();
    }
    if (t.isTSThisType(node)) {
        return t.thisTypeAnnotation();
    }
    if (t.isTSImportType(node)) {
        const exports = t.genericTypeAnnotation(t.identifier('$Exports'), t.typeParameterInstantiation([
            t.stringLiteralTypeAnnotation(node.argument.value),
        ]));
        if (node.qualifier) {
            if (t.isTSQualifiedName(node.qualifier)) {
                throw new Error('todo:');
            }
            const exportName = t.stringLiteralTypeAnnotation(node.qualifier.name);
            return t.genericTypeAnnotation(t.identifier('$PropertyType'), t.typeParameterInstantiation([exports, exportName]));
        }
        return exports;
    }
    if (t.isTSTypeQuery(node)) {
        if (t.isTSEntityName(node.exprName)) {
            return t.typeofTypeAnnotation(t.genericTypeAnnotation(convertTSEntityName_1.convertTSEntityName(node.exprName)));
        }
        if (t.isTSImportType(node.exprName)) {
            return t.genericTypeAnnotation(t.identifier('$Exports'), t.typeParameterInstantiation([
                t.stringLiteralTypeAnnotation(node.exprName.argument.value),
            ]));
        }
    }
    if (t.isTSTypeOperator(node)) {
        if (node.operator === 'typeof') {
            return t.typeofTypeAnnotation(convertTSType(node.typeAnnotation));
        }
        else if (node.operator === 'keyof') {
            return t.genericTypeAnnotation(t.identifier('$Keys'), t.typeParameterInstantiation([convertTSType(node.typeAnnotation)]));
        }
        else if (node.operator === 'unique' &&
            t.isTSSymbolKeyword(node.typeAnnotation)) {
            return t.genericTypeAnnotation(t.identifier('Symbol'));
        }
        else if (node.operator === 'readonly' &&
            t.isTSTupleType(node.typeAnnotation)) {
            return convertTSType(node.typeAnnotation);
        }
        else if (node.operator === 'readonly' &&
            t.isTSArrayType(node.typeAnnotation)) {
            return t.genericTypeAnnotation(t.identifier('$ReadOnlyArray'), t.typeParameterInstantiation([
                convertTSType(node.typeAnnotation.elementType),
            ]));
        }
        else {
            throw new Error('todo: not implemented');
        }
    }
    if (t.isTSUnionType(node)) {
        const tsTypes = node.types;
        return t.unionTypeAnnotation(tsTypes.map(v => {
            const tsType = convertTSType(v);
            // if (t.isFunctionTypeAnnotation(tsType)) {
            //   //tsParenthesizedType
            //   tsType = tsType;
            // }
            return Object.assign(Object.assign({}, tsType), baseNodeProps_1.baseNodeProps(v));
        }));
    }
    if (t.isTSVoidKeyword(node)) {
        return t.voidTypeAnnotation();
    }
    if (t.isTSFunctionType(node)) {
        const { typeParams, parameters, rest, returnType, } = convertFunctionTypeAnnotation_1.convertFunctionTypeAnnotation(node);
        return t.functionTypeAnnotation(typeParams, parameters, rest, returnType ? returnType : t.anyTypeAnnotation());
    }
    if (t.isTSTupleType(node)) {
        const elementTypes = node.elementTypes;
        return t.tupleTypeAnnotation(elementTypes.map(element => {
            if (t.isTSNamedTupleMember(element)) {
                return convertTSType(element.elementType);
            }
            else {
                return convertTSType(element);
            }
        }));
    }
    if (t.isTSObjectKeyword(node)) {
        return t.objectTypeAnnotation([], [], [], []);
    }
    if (t.isTSSymbolKeyword(node)) {
        return t.genericTypeAnnotation(t.identifier('Symbol'));
    }
    if (t.isTSOptionalType(node)) {
        return t.unionTypeAnnotation([
            convertTSType(node.typeAnnotation),
            t.voidTypeAnnotation(),
        ]);
    }
    if (t.isTSConditionalType(node)) {
        //todo: warn
        return t.anyTypeAnnotation();
    }
    throw new Error(`Unsupported flow type TSType(type=${node.type})`);
}
exports.convertTSType = convertTSType;
//# sourceMappingURL=convertTSType.js.map